/*
 * SimbaView.java
 */

package DTC;

import org.jdesktop.application.Action;
import org.jdesktop.application.ResourceMap;
import org.jdesktop.application.SingleFrameApplication;
import org.jdesktop.application.FrameView;
import org.jdesktop.application.Task;
import org.jdesktop.application.TaskMonitor;
import java.awt.event.ActionEvent;
import java.awt.event.ActionListener;
import javax.swing.Timer;
import javax.swing.Icon;
import javax.swing.JDialog;
import javax.swing.JFrame;
import java.util.Vector;
import javax.swing.JOptionPane;
import canbus.*;

/**
 * The application's main frame.
 */
public class DTCView extends FrameView {

    public DTCView(SingleFrameApplication app) {
        super(app);

        initComponents();

        // status bar initialization - message timeout, idle icon and busy animation, etc
        ResourceMap resourceMap = getResourceMap();
        int messageTimeout = resourceMap.getInteger("StatusBar.messageTimeout");
        messageTimer = new Timer(messageTimeout, new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                statusMessageLabel.setText("");
            }
        });
        messageTimer.setRepeats(false);
        int busyAnimationRate = resourceMap.getInteger("StatusBar.busyAnimationRate");
        for (int i = 0; i < busyIcons.length; i++) {
            busyIcons[i] = resourceMap.getIcon("StatusBar.busyIcons[" + i + "]");
        }
        busyIconTimer = new Timer(busyAnimationRate, new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                busyIconIndex = (busyIconIndex + 1) % busyIcons.length;
                statusAnimationLabel.setIcon(busyIcons[busyIconIndex]);
            }
        });
        idleIcon = resourceMap.getIcon("StatusBar.idleIcon");
        statusAnimationLabel.setIcon(idleIcon);
        progressBar.setVisible(false);

        // connecting action tasks to status bar via TaskMonitor
        TaskMonitor taskMonitor = new TaskMonitor(getApplication().getContext());
        taskMonitor.addPropertyChangeListener(new java.beans.PropertyChangeListener() {
            public void propertyChange(java.beans.PropertyChangeEvent evt) {
                String propertyName = evt.getPropertyName();
                if ("started".equals(propertyName)) {
                    if (!busyIconTimer.isRunning()) {
                        statusAnimationLabel.setIcon(busyIcons[0]);
                        busyIconIndex = 0;
                        busyIconTimer.start();
                    }
                    progressBar.setVisible(true);
                    progressBar.setIndeterminate(true);
                } else if ("done".equals(propertyName)) {
                    busyIconTimer.stop();
                    statusAnimationLabel.setIcon(idleIcon);
                    progressBar.setVisible(false);
                    progressBar.setValue(0);
                } else if ("message".equals(propertyName)) {
                    String text = (String)(evt.getNewValue());
                    statusMessageLabel.setText((text == null) ? "" : text);
                    messageTimer.restart();
                } else if ("progress".equals(propertyName)) {
                    int value = (Integer)(evt.getNewValue());
                    progressBar.setVisible(true);
                    progressBar.setIndeterminate(false);
                    progressBar.setValue(value);
                }
            }
        });
    }

    @Action
    public void showAboutBox() {
        if (aboutBox == null) {
            JFrame mainFrame = DTCApp.getApplication().getMainFrame();
            aboutBox = new DTCAboutBox(mainFrame);
            aboutBox.setLocationRelativeTo(mainFrame);
        }
        DTCApp.getApplication().show(aboutBox);
    }

    /** This method is called from within the constructor to
     * initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the Form Editor.
     */
    @SuppressWarnings("unchecked")
    // <editor-fold defaultstate="collapsed" desc="Generated Code">//GEN-BEGIN:initComponents
    private void initComponents() {

        mainPanel = new javax.swing.JPanel();
        jScrollPane1 = new javax.swing.JScrollPane();
        jList1 = new javax.swing.JList();
        searchButton = new javax.swing.JButton();
        checkRadio = new javax.swing.JRadioButton();
        clearRadio = new javax.swing.JRadioButton();
        menuBar = new javax.swing.JMenuBar();
        javax.swing.JMenu fileMenu = new javax.swing.JMenu();
        javax.swing.JMenuItem exitMenuItem = new javax.swing.JMenuItem();
        javax.swing.JMenu helpMenu = new javax.swing.JMenu();
        javax.swing.JMenuItem aboutMenuItem = new javax.swing.JMenuItem();
        statusPanel = new javax.swing.JPanel();
        javax.swing.JSeparator statusPanelSeparator = new javax.swing.JSeparator();
        statusMessageLabel = new javax.swing.JLabel();
        statusAnimationLabel = new javax.swing.JLabel();
        progressBar = new javax.swing.JProgressBar();
        buttonGroup1 = new javax.swing.ButtonGroup();

        mainPanel.setName("mainPanel"); // NOI18N

        jScrollPane1.setName("jScrollPane1"); // NOI18N

        jList1.setModel(new javax.swing.AbstractListModel() {
            String[] strings = { "Select an action and push the 'Run' button" };
            public int getSize() { return strings.length; }
            public Object getElementAt(int i) { return strings[i]; }
        });
        jList1.setName("jList1"); // NOI18N
        jScrollPane1.setViewportView(jList1);

        javax.swing.ActionMap actionMap = org.jdesktop.application.Application.getInstance(DTC.DTCApp.class).getContext().getActionMap(DTCView.class, this);
        searchButton.setAction(actionMap.get("runPushed")); // NOI18N
        searchButton.setName("searchButton"); // NOI18N

        buttonGroup1.add(checkRadio);
        checkRadio.setSelected(true);
        org.jdesktop.application.ResourceMap resourceMap = org.jdesktop.application.Application.getInstance(DTC.DTCApp.class).getContext().getResourceMap(DTCView.class);
        checkRadio.setText(resourceMap.getString("checkRadio.text")); // NOI18N
        checkRadio.setName("checkRadio"); // NOI18N

        buttonGroup1.add(clearRadio);
        clearRadio.setText(resourceMap.getString("clearRadio.text")); // NOI18N
        clearRadio.setName("clearRadio"); // NOI18N

        javax.swing.GroupLayout mainPanelLayout = new javax.swing.GroupLayout(mainPanel);
        mainPanel.setLayout(mainPanelLayout);
        mainPanelLayout.setHorizontalGroup(
            mainPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(mainPanelLayout.createSequentialGroup()
                .addGroup(mainPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                    .addGroup(mainPanelLayout.createSequentialGroup()
                        .addContainerGap()
                        .addComponent(jScrollPane1, javax.swing.GroupLayout.PREFERRED_SIZE, 417, javax.swing.GroupLayout.PREFERRED_SIZE))
                    .addGroup(mainPanelLayout.createSequentialGroup()
                        .addContainerGap()
                        .addComponent(searchButton, javax.swing.GroupLayout.DEFAULT_SIZE, 427, Short.MAX_VALUE))
                    .addGroup(mainPanelLayout.createSequentialGroup()
                        .addGap(171, 171, 171)
                        .addGroup(mainPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
                            .addComponent(clearRadio)
                            .addComponent(checkRadio))))
                .addContainerGap())
        );
        mainPanelLayout.setVerticalGroup(
            mainPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(mainPanelLayout.createSequentialGroup()
                .addContainerGap()
                .addComponent(jScrollPane1, javax.swing.GroupLayout.PREFERRED_SIZE, 182, javax.swing.GroupLayout.PREFERRED_SIZE)
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addComponent(checkRadio)
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addComponent(clearRadio)
                .addGap(10, 10, 10)
                .addComponent(searchButton)
                .addContainerGap(45, Short.MAX_VALUE))
        );

        menuBar.setName("menuBar"); // NOI18N

        fileMenu.setText(resourceMap.getString("fileMenu.text")); // NOI18N
        fileMenu.setName("fileMenu"); // NOI18N

        exitMenuItem.setAction(actionMap.get("quit")); // NOI18N
        exitMenuItem.setName("exitMenuItem"); // NOI18N
        fileMenu.add(exitMenuItem);

        menuBar.add(fileMenu);

        helpMenu.setText(resourceMap.getString("helpMenu.text")); // NOI18N
        helpMenu.setName("helpMenu"); // NOI18N

        aboutMenuItem.setAction(actionMap.get("showAboutBox")); // NOI18N
        aboutMenuItem.setName("aboutMenuItem"); // NOI18N
        helpMenu.add(aboutMenuItem);

        menuBar.add(helpMenu);

        statusPanel.setName("statusPanel"); // NOI18N

        statusPanelSeparator.setName("statusPanelSeparator"); // NOI18N

        statusMessageLabel.setName("statusMessageLabel"); // NOI18N

        statusAnimationLabel.setHorizontalAlignment(javax.swing.SwingConstants.LEFT);
        statusAnimationLabel.setName("statusAnimationLabel"); // NOI18N

        progressBar.setName("progressBar"); // NOI18N

        javax.swing.GroupLayout statusPanelLayout = new javax.swing.GroupLayout(statusPanel);
        statusPanel.setLayout(statusPanelLayout);
        statusPanelLayout.setHorizontalGroup(
            statusPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addComponent(statusPanelSeparator, javax.swing.GroupLayout.DEFAULT_SIZE, 447, Short.MAX_VALUE)
            .addGroup(statusPanelLayout.createSequentialGroup()
                .addContainerGap()
                .addComponent(statusMessageLabel)
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED, 277, Short.MAX_VALUE)
                .addComponent(progressBar, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE)
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED)
                .addComponent(statusAnimationLabel)
                .addContainerGap())
        );
        statusPanelLayout.setVerticalGroup(
            statusPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.LEADING)
            .addGroup(statusPanelLayout.createSequentialGroup()
                .addComponent(statusPanelSeparator, javax.swing.GroupLayout.PREFERRED_SIZE, 2, javax.swing.GroupLayout.PREFERRED_SIZE)
                .addPreferredGap(javax.swing.LayoutStyle.ComponentPlacement.RELATED, javax.swing.GroupLayout.DEFAULT_SIZE, Short.MAX_VALUE)
                .addGroup(statusPanelLayout.createParallelGroup(javax.swing.GroupLayout.Alignment.BASELINE)
                    .addComponent(statusMessageLabel)
                    .addComponent(statusAnimationLabel)
                    .addComponent(progressBar, javax.swing.GroupLayout.PREFERRED_SIZE, javax.swing.GroupLayout.DEFAULT_SIZE, javax.swing.GroupLayout.PREFERRED_SIZE))
                .addGap(3, 3, 3))
        );

        setComponent(mainPanel);
        setMenuBar(menuBar);
        setStatusBar(statusPanel);
    }// </editor-fold>//GEN-END:initComponents

    @Action
    public Task runPushed() {
        return new RunPushedTask(getApplication());
    }

    private class RunPushedTask extends org.jdesktop.application.Task<Object, Void> {
        private Vector<String> listEntries;
        private Timer stopTimer;
        
        RunPushedTask(org.jdesktop.application.Application app) {
            // Runs on the EDT.  Copy GUI state that
            // doInBackground() depends on from parameters
            // to RunPushedTask fields, here.
            super(app);

            // Initialize output list
            listEntries = new Vector<String>();

            // Update the output list
            jList1.setListData(listEntries);

            // Disable GUI elements
            searchButton.setEnabled(false);
        }

        @Override
        protected Object doInBackground() {

            CandapterPort candapterPort = new CandapterPort();
            CanbusMessage canMesg = null;
            boolean portFound = false;
            byte dtcBuf[] = new byte[2];
            int dtcType = 0;
            String listLine = "";

            // Attempt to open a CANdapter port
            portFound = candapterPort.openCanbus("");
            
            // If no port was found, exit
            if (!portFound) {
                // Print error to client
                JOptionPane.showMessageDialog(null,
                        "No CANdapters were found. Please make sure the CANdapter is plugged in.",
                        "CANdapter not found", JOptionPane.ERROR_MESSAGE);
                return null;
            }

            // Start polling CANBUS
            candapterPort.startPolling();
            
            // Start polling
            pollCanbus = true;

            // Timer to stop polling the CANBUS and end action
            stopTimer = new Timer(5000, new ActionListener() {

                public void actionPerformed(ActionEvent e) {
                    // We are done polling the CANBUS
                    pollCanbus = false;
                }
            });
            stopTimer.start();
            
            // Send DTC messages
            if (checkRadio.isSelected() == true) { // Check DTCs
                candapterPort.writeMessage(new CanbusMessage(0x7E2, 8, "0213B00000000000", false)); // HV-ECU DTCs
                candapterPort.writeMessage(new CanbusMessage(0x7E3, 8, "0213800000000000", false)); // Battery-ECU DTCs
                candapterPort.writeMessage(new CanbusMessage(0x7E0, 8, "0213B00000000000", false)); // Engine-ECU DTCs     
            } else { // Clear DTCs
                candapterPort.writeMessage(new CanbusMessage(0x7E2, 8, "0104000000000000", false)); // HV-ECU DTCs
                candapterPort.writeMessage(new CanbusMessage(0x7E3, 8, "0104000000000000", false)); // Battery-ECU DTCs
                candapterPort.writeMessage(new CanbusMessage(0x7E0, 8, "0104000000000000", false)); // Engine-ECU DTCs     
            }

            // Sleep for 500ms
//            try { Thread.sleep(500); } catch (Exception e) { }
            
            // Loop until stop timer triggers or stop button pressed
            while (pollCanbus == true) {

                // Read in a canbus message
                canMesg = candapterPort.readMessage();

                // If no message was available, sleep and try again                
                if (canMesg == null) {
                    try {
                        Thread.sleep(5);
                    } catch (Exception e) {}
                    continue;
                }
                
                // Clear list variable
                listLine = "";

                // Determine message ID
                switch (canMesg.getId()) {
                    case 0x7EB: // Battery ECU
                    case 0x7EA: // Hybrid ECU
                    case 0x7E8: // Engine ECU

                        // Check what type of message it is
                        if (canMesg.getByteAt(1) == 0x44 
                                && checkRadio.isSelected() == false) { // Clear DTCs
                            if(canMesg.getId() == 0x7EA)
                                listLine = "Hybid ECU DTCs cleared.";
                            else if(canMesg.getId() == 0x7EB)
                                listLine = "Battery ECU DTCs cleared.";
                            else if(canMesg.getId() == 0x7E8)
                                listLine = "Engine ECU DTCs cleared.";
                        }
                        else if (canMesg.getByteAt(1) == 0x53 
                                && canMesg.getByteAt(2) > 0 
                                && canMesg.getByteAt(0) <= 0x06
                                && checkRadio.isSelected() == true) {

                            // Loop through elements of message
                            for (int i = 1; i < canMesg.getByteAt(2) + 1 && i < 3; ++i) {
                                
                                // Read in DTC from canbus message object
                                dtcBuf[0] = canMesg.getByteAt((i * 2) + 1);
                                dtcBuf[1] = canMesg.getByteAt((i * 2) + 2);
                                
                                // Obtain type of DTC (Most significant 2 bits in DTC message)
                                dtcType = (dtcBuf[0] & 0xC0) >> 6;

                                // Remove 2 most significant bits in DTC message
                                dtcBuf[0] &= 0x3F;

                                // Debug message
//                              System.out.println("DTC TYPE: " + dtcType);                                
                                
                                // Translate DTC type into a character
                                switch (dtcType) {
                                    case 0x00:
                                        listLine = "P";
                                        break;
                                    case 0x01:
                                        listLine = "C";
                                        break;
                                    case 0x02:
                                        listLine = "B";
                                        break;
                                    case 0x03:
                                        listLine = "U";
                                        break;
                                    default:
                                        listLine = "?";
                                }
                                
                                // Build output line for display
                                listLine += "" + canMesg.getStrFrom(dtcBuf[0]) + canMesg.getStrFrom(dtcBuf[1]);
                            }
                        }
                        
                        // Add DTC entry to data display vector
                        if(!listLine.equals(""))
                            listEntries.add(listLine);

                        // Break out of switch case
                        break;
                }
            }

            // Stop polling
            candapterPort.stopPolling();
            
            // Close CANdapter port
            candapterPort.closeCanbus();
               
            // Stop timer
            stopTimer.stop();

            // Return result to succeeded method
            return null;
        }
        @Override protected void succeeded(Object result) {
            // Runs on the EDT.  Update the GUI based on
            // the result computed by doInBackground().
           
            // Check and make sure there is data in the list, otherwise add blank
            if(listEntries.size() <= 0) {
                if(checkRadio.isSelected() == true) 
                    listEntries.add("No DTCs found.");
                else
                    listEntries.add("No DTCs cleared.");
            }

            // Update the output list
            jList1.setListData(listEntries);
            
            // Enable GUI elements
            searchButton.setEnabled(true);
        }
    }

    // User added variable defines
    private boolean pollCanbus;
    
    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.ButtonGroup buttonGroup1;
    private javax.swing.JRadioButton checkRadio;
    private javax.swing.JRadioButton clearRadio;
    private javax.swing.JList jList1;
    private javax.swing.JScrollPane jScrollPane1;
    private javax.swing.JPanel mainPanel;
    private javax.swing.JMenuBar menuBar;
    private javax.swing.JProgressBar progressBar;
    private javax.swing.JButton searchButton;
    private javax.swing.JLabel statusAnimationLabel;
    private javax.swing.JLabel statusMessageLabel;
    private javax.swing.JPanel statusPanel;
    // End of variables declaration//GEN-END:variables

    private final Timer messageTimer;
    private final Timer busyIconTimer;
    private final Icon idleIcon;
    private final Icon[] busyIcons = new Icon[15];
    private int busyIconIndex = 0;

    private JDialog aboutBox;
}
