/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */

package canbus;

import java.nio.ByteBuffer;

/**
 *
 * @author andrew
 */
public class CanbusMessage {
    private int id;
    private int length;
    private ByteBuffer buffer;
    private boolean extended;
    private int timestamp;

    // Constructor for bytebuffer-based buffer
    public CanbusMessage(int id, int length, ByteBuffer buffer, boolean extended) {
        this.id = id;
        this.length = length;
        this.buffer = buffer;
        this.extended = extended;
        this.timestamp = -1;
    }
    
    // Constructor for string-based buffer
    public CanbusMessage(int id, int length, String message, boolean extended) {
        String chunk = "";
        int offset = 0, i = 0;
        ByteBuffer buffer = ByteBuffer.allocate(message.length() / 2);

        // Note: length is the first byte, ignore it
        for (i = 0; i < message.length() / 2; ++i) {
            chunk = message.substring(offset, offset + 2);
            offset += 2;
            buffer.put((byte) (Integer.parseInt(chunk, 16) & 0x000000FF));
        }

        // Rewind buffer
        buffer.rewind();

        // Assign message components
        this.id = id;
        this.length = length;
        this.buffer = buffer;
        this.extended = extended;
        this.timestamp = -1;
    }
    
    // Update length for CANBUS message
    public void setLength(int length) {
        this.length = length;
    }
    
    // Update timestamp for CANBUS message
    public void setTimestamp(int timestamp) {
        this.timestamp = timestamp;
    }
    
    // Update ID for CANBUS message
    public void setId(int id) {
        this.id = id;
    }
    
    // Update buffer for CANBUS message
    public void setBuffer(ByteBuffer buffer) {
        this.buffer = buffer;
    }
    
    // Update extended message boolean value
    public void setExtended(boolean extended) {
        this.extended = extended;
    }
    
    // Get the message buffer
    public ByteBuffer getBuffer() {
        return buffer;
    }
    
    // Get string based representation of message buffer
    public String getStrBuffer() {
        String retLine = "";
        String tmpLine = "";
        byte element = 0;
        
        // Convert byte buffer to string
        while(buffer.hasRemaining()) {
            element = buffer.get();    
            
            // Get line from element
            tmpLine = Integer.toString(element & 0xFF, 16).toUpperCase();

            // If value is less than 0x10, add a zero
            if(tmpLine.length() < 2)
                tmpLine = "0" + tmpLine;            

            // Add temporary line to return string
            retLine += tmpLine;
        }

        // Rewind buffer pointer so we can read again
        buffer.rewind();
        
        // Return string
        return retLine;
    }
    
    // Get timestamp for message
    public int getTimestamp() {
        return timestamp;
    }
    
    // Get specific byte from buffer
    public byte getByteAt(int index) {
        if(buffer.capacity() <= index || index < 0)
            return 0;
        else
            return buffer.get(index);
    }
    
    // Given a byte, return a hexadecimal string
    public String getStrFrom(byte element) {
        String retLine = "";
        
        retLine = Integer.toString(element & 0xFF, 16).toUpperCase();

        if(retLine.length() < 2)
            retLine = "0" + retLine;
        
        return retLine;
    }
    
    // Get string based representation of specific byte from buffer
    public String getStrAt(int index) {
        byte element = 0;
        String retLine = "";
        
        element = getByteAt(index);
        retLine = Integer.toString(element & 0xFF, 16).toUpperCase();

        if(retLine.length() < 2)
            retLine = "0" + retLine;
        
        return retLine;
    }

    // Get string based representation of specific byte from buffer
    public String getStrAt(int start, int stop) {
        byte element = 0;
        String retLine = "", chunk = "";
        
        // Loop through requested bytes
        for (int i = start; i < stop; i++) {
            // Get a byte
            element = getByteAt(start);
            
            // Split off one byte
            chunk = Integer.toString(element & 0xFF, 16).toUpperCase();

            // Make sure its at least two characters
            if (chunk.length() < 2) {
                chunk += "0" + retLine;
            }
            
            // Add chunk to returned string
            retLine += chunk;
        }
        
        // Return built string
        return retLine;
    }    
    
    // Get the hexadecimal string equivalent of id
    public String getStrId() {
        String retLine = "";
        
        retLine = Integer.toHexString(id).toUpperCase();
        
        if(extended == true) 
            while(retLine.length() < 8)
                retLine = "0" + retLine;
        else
            while(retLine.length() < 3)
                retLine = "0" + retLine;
        
        return retLine;
    }
    
    // Get the integer value of message id
    public int getId() {
        return id;
    }
    
    // Get the message length
    public int getLength() {
        return length;
    }
    
    // Get whether this is an extended message
    public boolean isExtended() {
        return extended;
    }
    
    // Return a string equivalent of message
    @Override
    public String toString() {
        String retVal = "";
        
        // Call string functions
        retVal = getStrId() + getLength() + getStrBuffer();
        
        // Return string
        return retVal;
    }
}
