/*
 * To change this template, choose Tools | Templates
 * and open the template in the editor.
 */
package canbus;

import java.io.*;
import java.util.*;
import gnu.io.*;
import java.nio.ByteBuffer;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

/**wri
 *
 * @author andrew
 */
public class CandapterPort implements CanbusPort {

    private CommPortIdentifier portId;
    private SerialPort serialPort;
    private OutputStream outputStream;
    private InputStream inputStream;
    private ArrayList<CanbusMessage> rxBuffer;
    private ArrayList<CanbusMessage> txBuffer;
    private Thread readerThread;
    private CanbusReader canbusReader;
    private Thread writerThread;
    private CanbusWriter canbusWriter;
    
    private boolean stopPolling = false;
    private boolean temp = false;
    private boolean debugging = false;    

    public boolean openCanbus(String defaultPort) throws java.lang.UnsatisfiedLinkError  {
        boolean portFound = false;
        Enumeration portList;

        // Otherwise find the real serial port
        portList = CommPortIdentifier.getPortIdentifiers();

        // If no port was supplied, pull one from active ports
        if(defaultPort.equals("")) {
            Vector<String> candapterPorts = getPorts();
            
            // Make sure we have an element
            if(candapterPorts.size() > 0) {
                defaultPort = candapterPorts.get(0);
            }
        }
        
        // Loop through all serial/parallel ports found
        while (portList.hasMoreElements()) {
            portId = (CommPortIdentifier) portList.nextElement();

            // Give default port prescedence if it is specified
            if (!(defaultPort.equals(portId.getName()))) {
                continue;
            }

            // Only poll serial ports
            if (portId.getPortType() == CommPortIdentifier.PORT_SERIAL) {

                // Attempt to open the serial port
                try {
                    serialPort = (SerialPort) portId.open("CANdapter API", 50);
                } catch (PortInUseException e) {
                    // If port is in use, continue to next port
                    continue;
                }

                // Enable timeout for RX
                try {
                    serialPort.enableReceiveTimeout(100);
                    serialPort.enableReceiveThreshold(0);
                } catch (UnsupportedCommOperationException e) {
                    System.err.println(e);
                    e.printStackTrace();
                }

                try {
                    // Create streams and parsers
                    outputStream = serialPort.getOutputStream();
                    inputStream = serialPort.getInputStream();

                    // Set longer timeout
                    try {
// CHANGED: WAS 1000 (1second)
                        serialPort.enableReceiveTimeout(100000);
                    } catch (UnsupportedCommOperationException e) {
                        System.err.println(e);
                        e.printStackTrace();
                    }

                    // Set port found flag
                    System.out.println("Found port: " + portId.getName());
                    portFound = true;

                } catch (IOException e) {
                    // Display error and clear found flag
                    System.err.println(e);
                    portFound = false;

                    // Close serial port if it was opened
                    if (serialPort != null) {
                        serialPort.close();
                    }

                    // Close other ports
                    try {
                        if (inputStream != null) {
                            inputStream.close();
                        }
                        if (outputStream != null) {
                            outputStream.close();
                        }
 //                       if (inputReader != null) {
 //                           inputReader.close();
 //                       }
                    } catch (IOException io) {
                    }

                    // Continue to next serial port
                    continue;
                }

                // Legacy support for serial communications
                try {
                    serialPort.setSerialPortParams(921600, SerialPort.DATABITS_8, SerialPort.STOPBITS_1, SerialPort.PARITY_NONE);
                    serialPort.notifyOnOutputEmpty(true);
                } catch (UnsupportedCommOperationException e) {
                } catch (Exception e) {
                    System.out.println(e.toString());
                    continue;
                }
            }
        }

       if(defaultPort.equals("DEBUG")) {
            try {
System.out.println("DEBUGGING");
                inputStream = new FileInputStream("sampledata.txt");
                portFound = true;
                debugging = true;
            } catch (FileNotFoundException ex) {
                ex.printStackTrace();
            }
        }        
                
        // Return whether or not we found a valid port
        return portFound;
    }

    // Open a file as if it were a CANdapter
    public boolean openFile(String filePath) {
        boolean fileFound = false;

        try {
            inputStream = new FileInputStream(filePath);
            fileFound = true;
            debugging = true;
        } catch (FileNotFoundException ex) {
            ex.printStackTrace();
        }      
                
        // Return whether or not we found a valid file
        return fileFound;
    }    
    
    // Close the CANBUS and terminate reading and writing threads
    synchronized public void closeCanbus() {
        String messageString = "";
               
        // Stop polling
        stopPolling();
                
        // Close CANdapter port
        if(outputStream != null) {
            try {
                messageString = "\015C\015";
                outputStream.write(messageString.getBytes());
            } catch (Exception e) {
                System.err.println(e);
                e.printStackTrace();                
            }
        }

        // Close open ports
        try {            
            if (inputStream != null) {
                inputStream.close();
            }
            if (outputStream != null) {
                outputStream.close();
            }
            
            if (serialPort != null) {
                serialPort.close();
            }
            
        } catch (Exception e) {
            System.err.println(e);
            e.printStackTrace();
        }

        return;
    }
    
    // Returns debugging state of CANdapter
    public boolean isDebugging() {
        return debugging;
    }

    // Get CANdapter version number
    public String getVersionNumber() {
        BufferedReader inputReader = null;
        StreamTokenizer inputTokenizer = null;
        String versionNumber = "";
        boolean versionFound = false;
        int major = 0, minor = 0, patch = 0;
        
        try {
            // Conform to deprecated API for StreamTokenizer
            inputReader = new BufferedReader(new InputStreamReader(inputStream));
            inputTokenizer = new StreamTokenizer(inputReader);
            String patternStr = "[V]([0-9]+)";
            Pattern pattern = Pattern.compile(patternStr);
            Matcher matcher;

            // Setup tokenizer to split based on 't' fields
            inputTokenizer.parseNumbers();
            inputTokenizer.wordChars('V', 'V');
            inputTokenizer.eolIsSignificant(false);

            // Loop until we find data
            do {
                // Request serial number
                outputStream.write("V\015".getBytes());

                String lineIn = inputReader.readLine();
                matcher = pattern.matcher(lineIn);
System.out.println("LINE: " + lineIn);
                // Check if REGEX appears in text
                if (matcher.find() == true) {
                    versionNumber = matcher.group(1);
                    versionFound = true;
                    major = Integer.parseInt(versionNumber.substring(0, 2), 16);
                    minor = Integer.parseInt(versionNumber.substring(2, 4), 16);
                    patch = Integer.parseInt(versionNumber.substring(4, 6), 16);
                    break;
                }
            } while (versionFound == false);
        } catch (Exception e) {
            e.printStackTrace();
        }
        
        return major + "." + minor + "." + patch;
    }    
    
    // Get CANdapter serial number
    public String getSerialNumber() {
        BufferedReader inputReader = null;
        StreamTokenizer inputTokenizer = null;
        String serialNumber = "";
        boolean serialFound = false;
        
        try {
            // Conform to deprecated API for StreamTokenizer
            inputReader = new BufferedReader(new InputStreamReader(inputStream));
            inputTokenizer = new StreamTokenizer(inputReader);
            String patternStr = "[N](.+)";
            Pattern pattern = Pattern.compile(patternStr);
            Matcher matcher;

            // Setup tokenizer to split based on 't' fields
            inputTokenizer.parseNumbers();
            inputTokenizer.wordChars('N', 'N');
            inputTokenizer.eolIsSignificant(false);

            // Loop until we find data
            do {
                // Request serial number
                outputStream.write("N\015".getBytes());

                String lineIn = inputReader.readLine();
                matcher = pattern.matcher(lineIn);
System.out.println("LINE: " + lineIn);
                // Check if REGEX appears in text
                if (matcher.find() == true) {
                    serialNumber = matcher.group(1);
                    serialFound = true;
                    break;
                }
            } while (serialFound == false);
        } catch (Exception e) {
            e.printStackTrace();
        }
        
        return serialNumber;
    }    
    
    // Return a vector of all CANdapter ports found
    public static Vector<String> getPorts() {
        Enumeration thePorts = CommPortIdentifier.getPortIdentifiers();
        Vector<String> serialPorts = new Vector<String>();
        boolean portFound = false;
        CommPort thePort = null;
                    
        while (thePorts.hasMoreElements()) {
            CommPortIdentifier com = (CommPortIdentifier) thePorts.nextElement();

            switch (com.getPortType()) {
                case CommPortIdentifier.PORT_SERIAL:
                    BufferedReader inputReader = null;
                    StreamTokenizer inputTokenizer = null;
                    
                    try {
                        thePort = com.open("CANdapter API", 50);
                        String command = "V\015";

                        // Enable timeout for RX
                        thePort.enableReceiveTimeout(100);
                        thePort.enableReceiveThreshold(0);
                        
                        // Verify that this is a CANdapter
                        thePort.getOutputStream().write(command.getBytes());
                        
                        // Conform to deprecated API for StreamTokenizer
                        inputReader = new BufferedReader(new InputStreamReader(thePort.getInputStream()));
                        inputTokenizer = new StreamTokenizer(inputReader);
                        String patternStr = "([V][0-9][0-9][0-9][0-9][0-9][0-9])";
                        Pattern pattern = Pattern.compile(patternStr);
                        Matcher matcher;
                        
                        // Setup tokenizer to split based on 't' fields
                        inputTokenizer.parseNumbers();
                        inputTokenizer.wordChars('V', 'V');
                        inputTokenizer.eolIsSignificant(false);

                        // Loop until we find data
                        do {
                            String lineIn = inputReader.readLine();
                            matcher = pattern.matcher(lineIn);
                            
                            // Check if REGEX appears in text
                            if(matcher.find() == true) {
                                serialPorts.add(com.getName());
                                portFound = true;
System.out.println("COMM PORT FOUND: " + com.getName());                                
                                break;
                            }
                        } while(portFound == false);

                        // Close the port
                        thePort.close();
                    } catch (Exception e) {
                        if(thePort != null)
                            thePort.close();
                    }
            }
        }   
        
        return serialPorts;
    }   
    
    // Set timestamp
    public void setTimestamp(boolean timestamp) {
        String messageString;
        
        // Send command to CANdapter
        if(outputStream != null) {
            try {
                messageString = "\015A" + (timestamp == true ? 1 : 0) + "\015";
                outputStream.write(messageString.getBytes());
            } catch (Exception e) {
                System.err.println(e);
                e.printStackTrace();                
            }        
        }
    }

    // Set filter
    public void setFilter(int filterNum, int maskNum, int filterValue, boolean extended) {
        String messageString;
        
        // Send command to CANdapter
        if(outputStream != null) {
            try {
                messageString = "\015F" + (extended ? "1" : "0");
                messageString += Integer.toString(filterNum);
                messageString += Integer.toString(maskNum);
                if(extended)
                    messageString += String.format("%08x", filterValue);
                else
                    messageString += String.format("%03x", filterValue);                    
                messageString += "\015";
                outputStream.write(messageString.getBytes());
            } catch (Exception e) {
                System.err.println(e);
                e.printStackTrace();                
            }   
        }
    }    

    // Set mask
    public void setMask(int maskNum, int maskValue, boolean extended) {
        String messageString;
        
        // Send command to CANdapter
        if(outputStream != null) {
            try {
                messageString = "\015M" + (extended ? "1" : "0");
                messageString += Integer.toString(maskNum);
                if(extended)
                    messageString += String.format("%08x", maskValue);
                else
                    messageString += String.format("%03x", maskValue);                    
                messageString += "\015";
                outputStream.write(messageString.getBytes());
            } catch (Exception e) {
                System.err.println(e);
                e.printStackTrace();                
            }  
        }
    }    

    // Set baudrate
    public void setBaudrate(int baudrate) {
        String messageString;

        // Enforce maximum values
        if(baudrate < 0)
            baudrate = 0;
        if(baudrate > 8)
            baudrate = 8;
        
        // Send command to CANdapter
        if(outputStream != null) {        
            try {
                messageString = "\015S" + baudrate + "\015";
                outputStream.write(messageString.getBytes());
            } catch (Exception e) {
                System.err.println(e);
                e.printStackTrace();                
            }     
        }
    }     
    
    public void writeMessage(CanbusMessage canbusMessage) {
        /* Proceedure:
         * 1) Write command character (based on extended boolean)
         * 2) Write ID (based on extended boolean)
         * 3) Write message length
         * 4) Write message body
         * 
         * EXCEPTION: Catch an IOException and attempt to re-open port if discovered
         */

        transmitMessage(canbusMessage);
        return;
    }

    public CanbusMessage readMessage() {
        /* Proceedure:
         * 1) Read in first character
         * 2) Read in ID (length based on character received)
         * 3) Read in message length
         * 4) Read in each byte based on message length
         * 5) Convert each byte to int values and store in bytebuffer
         * 6) Return new CanbusMessage object
         *
         * EXCEPTION: Catch an IOException and attempt to re-open port if discovered         
         */

        // We didn't get a message, return null
        return receiveMessage();
    }

    public boolean isTxEmpty() {
        return txBuffer.isEmpty();
    }
    
    public int getRxSize() {
        return rxBuffer.size();
    }
    
    public boolean isRxEmpty() {
        return rxBuffer.isEmpty();
    }
    
    synchronized public boolean transmitMessage(CanbusMessage message) {
        if(txBuffer.size() < 128) {
            txBuffer.add(message);

//            try {Thread.sleep(5); } catch (Exception e) { }
//System.out.println("TX(" + txBuffer.size() + ": " + txBuffer.get(txBuffer.size()-1));            
            return true;
        } 
        return false;
    }
    
    synchronized public CanbusMessage receiveMessage() {
        if(rxBuffer.size() > 0) {
            return rxBuffer.remove(0);
        }
        
        return null;
    }
    
    public void startPolling() {

        // Send "OPEN" command to CANdapter to start polling
        if(outputStream != null) {
            try {
                String messageString = "";

                // Clear message buffer
                messageString = "\015\015\015";
                outputStream.write(messageString.getBytes());

                // Open the CAN channel
                messageString = "O\015";
                outputStream.write(messageString.getBytes());

            } catch (IOException e) {
                e.printStackTrace();
            }
        }
        
        // Initialize buffers
        rxBuffer = new ArrayList<CanbusMessage>();
        txBuffer = new ArrayList<CanbusMessage>();
        
        // Clear stop boolean
        stopPolling = false;
        
        // Start threads
        if(inputStream != null) {
            canbusReader = new CanbusReader(inputStream);
            readerThread = (new Thread(canbusReader));
        }
        
        if(outputStream != null) {
            canbusWriter = new CanbusWriter(outputStream);
            writerThread = (new Thread(canbusWriter));
        }
            
        // Start threads
        if(inputStream != null)
            readerThread.start();
        if (outputStream != null)
            writerThread.start();
    }
    
    public void stopPolling() {
        // Set stop boolean
        stopPolling = true;

        // Terminate read thread
        readerThread.interrupt();
        canbusReader.stopReading();
    }
    
    private class CanbusReader implements Runnable {
        private InputStream in;
        private StreamTokenizer inputTokenizer;
        private BufferedReader inputReader;
        private final String validHexChars = "0123456789ABCDEFabcdef";
        private boolean isNumber = false;
        private boolean debugging = false;
        private int offset = 0;
        private int timestamp = -1, prevTimestamp = -1, diffTimestamp = 0;
        private String sid = "", mesg = "", length = "", chunk = "";
        
        public CanbusReader (InputStream in) {
            this.in = in;
            
            // If inputstream is a file, we are debugging
            if(in instanceof FileInputStream)
                debugging = true;
            else
                debugging = false;

            // Conform to deprecated API for StreamTokenizer
            inputReader = new BufferedReader(new InputStreamReader(inputStream));
            inputTokenizer = new StreamTokenizer(inputReader);

            // Setup tokenizer to split based on 't' fields
            inputTokenizer.parseNumbers();
            inputTokenizer.wordChars('t', 't');
            inputTokenizer.wordChars('x', 'x');
            inputTokenizer.eolIsSignificant(false);  
        }
        
        void stopReading() {
            System.out.println("STOPPING READING....?");
            stopPolling = true;
        }
        
        synchronized public void enqueue(CanbusMessage message) {
//System.out.println("READ: t" + message.toString());
            rxBuffer.add(message);
        }       
        
        public void run () {
            CanbusMessage newMesg = null;

            // Add runtime hook to close CANBUS
            Runtime.getRuntime().addShutdownHook(new Thread() {
                public void run() {
                    System.out.println("Running READER shutdown hook.");

                    // Save variablehandler state to file
                    try {                   
                        closeCanbus();
                    } catch (Exception e) {
                        System.err.println(e);
                        e.printStackTrace();
                    }             
                }
            });            
            
            // Enable short timeout so we can stop whenever we want
            try {
                if(serialPort != null)
                    serialPort.enableReceiveTimeout(100);
            } catch (Exception e) {
                System.err.println(e);
                e.printStackTrace();
            }
            
            // Loop until boolean is set
            while(stopPolling == false) {
                try {
/*
                    // Only start polling if there is active data
                    if(inputStream.available() < 5) {
                        try {Thread.sleep(5); } catch (Exception e) { }
                        continue;
                    }
*/
                    // If we are debugging and buffer is filling up, let harvester catch up
                    if(debugging == true) {
                        if(rxBuffer.size() > 100) {
                            try { Thread.sleep(50); } catch (Exception e) { }
                        }
                    }
                    
                    if (inputTokenizer.nextToken() != StreamTokenizer.TT_EOF) {
                        // Initialize boolean
                        isNumber = true;      

                        // If the string value is null, go to next token
                        if (inputTokenizer.sval == null || inputTokenizer.sval.length() == 0) {
                            continue;
                        }

                        // Loop through value from tokenizer
                        for (int i = 1; i < inputTokenizer.sval.length(); ++i) {
                            // If character is invalid, mark as invalid
                            if (validHexChars.indexOf(inputTokenizer.sval.charAt(i)) == -1) {
                                isNumber = false;
                                break;
                            }
                        }

                        // Continue if there were invalid characters or length
                        if (isNumber == false || inputTokenizer.sval.length() <= 4 || inputTokenizer.sval.length() % 2 == 0) {
                            continue;
                        }

                        sid = (inputTokenizer.sval).substring(1, 4);
                        length = (inputTokenizer.sval).substring(4, 5);
                        mesg = (inputTokenizer.sval).substring(5);

                        ByteBuffer buffer = ByteBuffer.allocate(mesg.length()/2);
                        chunk = "";
                        offset = 0;

                        // Note: length is the first byte, ignore it
                        for (int i = 0; i < mesg.length() / 2; ++i) {
                            chunk = mesg.substring(offset, offset + 2);
                            offset += 2;
                            buffer.put((byte) (Integer.parseInt(chunk, 16) & 0x000000FF));
                        }

                        // Rewind buffer
                        buffer.rewind();

//System.out.println("ID: " + sid + " LEN: " + length + " BUF: " + buffer + "");                        

                        // Only add to message queue if length == buffer
                        if(Integer.parseInt(length, 16) <= buffer.capacity()) {
                            // Add new CANBUS message to list
                            newMesg = new CanbusMessage(Integer.parseInt(sid, 16),
                                    Integer.parseInt(length, 16),
                                    buffer,
                                    inputTokenizer.sval.charAt(0) == 'x');
                            enqueue(newMesg);
                        }

                        // Set timestamp (if available)
                        if(newMesg != null && mesg.length() > Integer.parseInt(length, 16) * 2) {
                            try {
                                timestamp = Integer.parseInt(mesg.substring(Integer.parseInt(length)*2, Integer.parseInt(length)*2 + 4), 16);
                                newMesg.setTimestamp(timestamp);
                            } catch (Exception e) {
                                System.err.println(e);
                                e.printStackTrace();
                            }
                        }
                    }
                } catch (IOException ex) {
                    System.out.println("IO Exception Ocurred: " + ex);
  //                  ex.printStackTrace();
                } catch (Exception e) { 
                    System.err.println(e);
                    e.printStackTrace();
                }
            }
            
System.out.println("STOPPING READ!");            
        }
    }     
    
    private class CanbusWriter implements Runnable {
        OutputStream out;
        
        public CanbusWriter (OutputStream out) {
            this.out = out;
        }

        // Dequeue a message from the transmit buffer
        synchronized public CanbusMessage dequeue() {
            CanbusMessage returnMesg;
                
            // Only remove a message if it is big enough
            if(txBuffer.size() > 0) {
                returnMesg = txBuffer.remove(0);
                return returnMesg;
            }
            return null;
        }             
        
        public void run () {
            String outputMessage = "";
            CanbusMessage canbusMessage = null;
            
            // Loop until boolean is set
            while(stopPolling == false) {
                if(txBuffer.isEmpty() == false) {
                    // Get data
                    canbusMessage = dequeue();

                    if(canbusMessage == null)
                        continue;
                    
                    outputMessage = canbusMessage.toString();
/*
System.out.println("OUT: " + canbusMessage);                    
System.out.println("ID: " + canbusMessage.getStrId());
System.out.println("Buffer: " + canbusMessage.getStrBuffer());
*/
                    // Append transmission character
                    if (canbusMessage.isExtended() == true) {
                        outputMessage = "x" + outputMessage;
                    } else {
                        outputMessage = "t" + outputMessage;
                    }
                    try {
System.out.println("MESG TO BCAST: " + outputMessage + ".");
                        outputMessage += "\015" + "O" + "\015";
                        outputStream.write(outputMessage.getBytes());
                    } catch (Exception e) {
                        e.printStackTrace();
                    }
                } else {
                    try {
                        Thread.sleep(5);
                    } catch (InterruptedException ex) {
                        ex.printStackTrace();
                    }
                }
            }            
        }
    }         
}
